@@ -2,6 +2,7 @@
 
 /* 
  * Copyright (C) 2001 Havoc Pennington, Copyright (C) 2002 Red Hat Inc.
+ * Copyright (C) 2006 Elijah Newren
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -41,6 +42,7 @@
  */
 #define KEY_MOUSE_BUTTON_MODS "/apps/metacity/general/mouse_button_modifier"
 #define KEY_FOCUS_MODE "/apps/metacity/general/focus_mode"
+#define KEY_RAISE_ON_CLICK "/apps/metacity/general/raise_on_click"
 #define KEY_ACTION_DOUBLE_CLICK_TITLEBAR "/apps/metacity/general/action_double_click_titlebar"
 #define KEY_AUTO_RAISE "/apps/metacity/general/auto_raise"
 #define KEY_AUTO_RAISE_DELAY "/apps/metacity/general/auto_raise_delay"
@@ -76,26 +78,27 @@ static guint changed_idle;
 #endif
 static GList *listeners = NULL;
 
-static gboolean use_system_font = TRUE;
+static gboolean use_system_font = FALSE;
 static PangoFontDescription *titlebar_font = NULL;
 static MetaVirtualModifier mouse_button_mods = Mod1Mask;
 static MetaFocusMode focus_mode = META_FOCUS_MODE_CLICK;
+static gboolean raise_on_click = TRUE;
 static char* current_theme = NULL;
 static int num_workspaces = 4;
 static MetaActionDoubleClickTitlebar action_double_click_titlebar =
-    META_ACTION_DOUBLE_CLICK_TITLEBAR_TOGGLE_SHADE;
+    META_ACTION_DOUBLE_CLICK_TITLEBAR_TOGGLE_MAXIMIZE;
 static gboolean application_based = FALSE;
 static gboolean disable_workarounds = FALSE;
 static gboolean auto_raise = FALSE;
 static gboolean auto_raise_delay = 500;
-static gboolean provide_visual_bell = TRUE;
+static gboolean provide_visual_bell = FALSE;
 static gboolean bell_is_audible = TRUE;
 static gboolean reduced_resources = FALSE;
 static gboolean gnome_accessibility = FALSE;
 static char *cursor_theme = NULL;
 static int   cursor_size = 24;
 
-static MetaVisualBellType visual_bell_type = META_VISUAL_BELL_INVALID;
+static MetaVisualBellType visual_bell_type = META_VISUAL_BELL_FULLSCREEN_FLASH;
 static MetaButtonLayout button_layout = {
   {
     META_BUTTON_FUNCTION_MENU,
@@ -119,10 +122,11 @@ static char *terminal_command = NULL;
 static char *workspace_names[MAX_REASONABLE_WORKSPACES] = { NULL, };
 
 #ifdef HAVE_GCONF
-static gboolean update_use_system_font   (gboolean    value);
+static gboolean update_use_system_font    (gboolean    value);
 static gboolean update_titlebar_font      (const char *value);
 static gboolean update_mouse_button_mods  (const char *value);
 static gboolean update_focus_mode         (const char *value);
+static gboolean update_raise_on_click     (gboolean    value);
 static gboolean update_theme              (const char *value);
 static gboolean update_visual_bell        (gboolean v1, gboolean v2);
 static gboolean update_visual_bell_type   (const char *value);
@@ -137,8 +141,6 @@ static gboolean update_window_binding     (const char *name,
                                            const char *value);
 static gboolean update_screen_binding     (const char *name,
                                            const char *value);
-static gboolean update_binding            (MetaKeyPref *binding,
-                                           const char  *value);
 static gboolean update_command            (const char  *name,
                                            const char  *value);
 static gboolean update_terminal_command   (const char *value);
@@ -158,6 +160,8 @@ static char* gconf_key_for_workspace_name (int i);
 
 static void queue_changed (MetaPreference  pref);
 #endif /* HAVE_GCONF */
+static gboolean update_binding            (MetaKeyPref *binding,
+                                           const char  *value);
 
 static void     init_bindings             (void);
 static void     init_commands             (void);
@@ -297,6 +301,29 @@ cleanup_error (GError **error)
       *error = NULL;
     }
 }
+
+/* get_bool returns TRUE if *val is filled in, FALSE otherwise */
+static gboolean
+get_bool (const char *key, gboolean *val)
+{
+  GError     *err = NULL;
+  GConfValue *value;
+  gboolean    filled_in = FALSE;
+
+  value = gconf_client_get (default_client, key, &err);
+  cleanup_error (&err);
+  if (value)
+    {
+      if (value->type == GCONF_VALUE_BOOL)
+        {
+          *val = gconf_value_get_bool (value);
+          filled_in = TRUE;
+        }
+      gconf_value_free (value);
+    }
+
+  return filled_in;
+}
 #endif /* HAVE_GCONF */
 
 void
@@ -346,6 +373,9 @@ meta_prefs_init (void)
   update_focus_mode (str_val);
   g_free (str_val);  
 
+  if (get_bool (KEY_RAISE_ON_CLICK, &bool_val))
+    update_raise_on_click (bool_val);
+  
   str_val = gconf_client_get_string (default_client,
                                      KEY_ACTION_DOUBLE_CLICK_TITLEBAR,
 				     &err);
@@ -353,11 +383,15 @@ meta_prefs_init (void)
   update_action_double_click_titlebar (str_val);
   g_free (str_val);
 
-  bool_val = gconf_client_get_bool (default_client, KEY_AUTO_RAISE,
-				    &err);
-  cleanup_error (&err);
-  update_auto_raise (bool_val);
+  if (get_bool (KEY_AUTO_RAISE, &bool_val))
+    update_auto_raise (bool_val);
 
+  /* FIXME: Since auto_raise_delay of 0 is valid and gconf_client_get_int
+   * silently returns that value when KEY_AUTO_RAISE_DELAY doesn't exist,
+   * we should be using gconf_client_get() instead if we cared about
+   * careful error checking of the key-doesn't-exist case.  Since this
+   * setting is crap, though, I'm adding a FIXME instead of fixing it.  ;-)
+   */
   int_val = gconf_client_get_int (default_client, KEY_AUTO_RAISE_DELAY,
 				  &err);
   cleanup_error (&err);
@@ -375,10 +409,8 @@ meta_prefs_init (void)
    * just lazy. But they keys ought to be set, anyhow.
    */
   
-  bool_val = gconf_client_get_bool (default_client, KEY_USE_SYSTEM_FONT,
-                                    &err);
-  cleanup_error (&err);
-  update_use_system_font (bool_val);  
+  if (get_bool (KEY_USE_SYSTEM_FONT, &bool_val))
+    update_use_system_font (bool_val);
   
   str_val = gconf_client_get_string (default_client, KEY_TITLEBAR_FONT,
                                      &err);
@@ -391,15 +423,11 @@ meta_prefs_init (void)
   cleanup_error (&err);
   update_num_workspaces (int_val);
 
-  bool_val = gconf_client_get_bool (default_client, KEY_APPLICATION_BASED,
-                                    &err);
-  cleanup_error (&err);
-  update_application_based (bool_val);
+  if (get_bool (KEY_APPLICATION_BASED, &bool_val))
+    update_application_based (bool_val);
 
-  bool_val = gconf_client_get_bool (default_client, KEY_DISABLE_WORKAROUNDS,
-                                    &err);
-  cleanup_error (&err);
-  update_disable_workarounds (bool_val);
+  if (get_bool (KEY_DISABLE_WORKAROUNDS, &bool_val))
+    update_disable_workarounds (bool_val);
 
   str_val = gconf_client_get_string (default_client, KEY_BUTTON_LAYOUT,
                                      &err);
@@ -407,13 +435,9 @@ meta_prefs_init (void)
   update_button_layout (str_val);
   g_free (str_val);
   
-  bool_val = gconf_client_get_bool (default_client, KEY_VISUAL_BELL,
-                                    &err);
-  cleanup_error (&err);
-  bool_val_2 = gconf_client_get_bool (default_client, KEY_AUDIBLE_BELL,
-                                    &err);
-  cleanup_error (&err);
-  update_visual_bell (bool_val, bool_val_2);
+  if (get_bool (KEY_VISUAL_BELL,  &bool_val) ||
+      get_bool (KEY_AUDIBLE_BELL, &bool_val_2))
+    update_visual_bell (bool_val, bool_val_2);
 
   str_val = gconf_client_get_string (default_client, KEY_VISUAL_BELL_TYPE,
                                      &err);
@@ -421,10 +445,8 @@ meta_prefs_init (void)
   update_visual_bell_type (str_val);
   g_free (str_val);
 
-  bool_val = gconf_client_get_bool (default_client, KEY_REDUCED_RESOURCES,
-				    &err);
-  cleanup_error (&err);
-  update_reduced_resources (bool_val);
+  if (get_bool (KEY_REDUCED_RESOURCES, &bool_val))
+    update_reduced_resources (bool_val);
 
   str_val = gconf_client_get_string (default_client, KEY_TERMINAL_COMMAND,
                                      &err);
@@ -432,11 +454,8 @@ meta_prefs_init (void)
   update_terminal_command (str_val);
   g_free (str_val);
 
-  bool_val = gconf_client_get_bool (default_client, KEY_GNOME_ACCESSIBILITY,
-                                    &err);
-  
-  cleanup_error (&err);
-  update_gnome_accessibility (bool_val);
+  if (get_bool (KEY_GNOME_ACCESSIBILITY, &bool_val))
+    update_gnome_accessibility (bool_val);
 
   str_val = gconf_client_get_string (default_client, KEY_CURSOR_THEME,
                                      &err);
@@ -448,6 +467,14 @@ meta_prefs_init (void)
                                   &err);
   cleanup_error (&err);
   update_cursor_size (int_val);
+#else  /* HAVE_GCONF */
+  /* Set defaults for some values that can't be set at initialization time of
+   * the static globals.  In the case of the theme, note that there is code
+   * elsewhere that will do everything possible to fallback to an existing theme
+   * if the one here does not exist.
+   */
+  titlebar_font = pango_font_description_from_string ("Sans Bold 10");
+  current_theme = g_strdup ("Atlanta");
 #endif /* HAVE_GCONF */
   
   /* Load keybindings prefs */
@@ -551,6 +578,22 @@ change_notify (GConfClient    *client,
       if (update_focus_mode (str))
         queue_changed (META_PREF_FOCUS_MODE);
     }
+  else if (strcmp (key, KEY_RAISE_ON_CLICK) == 0)
+    {
+      gboolean b;
+
+      if (value && value->type != GCONF_VALUE_BOOL)
+        {
+          meta_warning (_("GConf key \"%s\" is set to an invalid type\n"),
+                        KEY_RAISE_ON_CLICK);
+          goto out;
+        }
+
+      b = value ? gconf_value_get_bool (value) : TRUE;
+      
+      if (update_raise_on_click (b))
+        queue_changed (META_PREF_RAISE_ON_CLICK);
+    }
   else if (strcmp (key, KEY_THEME) == 0)
     {
       const char *str;
@@ -972,6 +1015,18 @@ update_focus_mode (const char *value)
 }
 #endif /* HAVE_GCONF */
 
+#ifdef HAVE_GCONF
+static gboolean
+update_raise_on_click (gboolean value)
+{
+  gboolean old = raise_on_click;
+
+  raise_on_click = value;
+
+  return old != value;
+}
+#endif /* HAVE_GCONF */
+
 #ifdef HAVE_GCONF
 static gboolean
 update_theme (const char *value)
@@ -1018,6 +1073,15 @@ meta_prefs_get_focus_mode (void)
   return focus_mode;
 }
 
+gboolean
+meta_prefs_get_raise_on_click (void)
+{
+  /* Force raise_on_click on for click-to-focus, as requested by Havoc
+   * in #326156.
+   */
+  return raise_on_click || focus_mode == META_FOCUS_MODE_CLICK;
+}
+
 const char*
 meta_prefs_get_theme (void)
 {
@@ -1063,7 +1127,11 @@ update_cursor_size (int value)
 {
   int old = cursor_size;
 
-  cursor_size = value;
+  if (value >= 1 && value <= 128)
+    cursor_size = value;
+  else
+    meta_warning (_("%d stored in GConf key %s is not a reasonable cursor_size; must be in the range 1..128\n"),
+                    value, KEY_CURSOR_SIZE);
 
   return old != value;
 }
@@ -1113,7 +1181,6 @@ update_visual_bell_type (const char *value)
 
   return (visual_bell_type != old_bell_type);
 }
-#endif /* HAVE_GCONF */
 
 static gboolean
 update_visual_bell (gboolean visual_bell, gboolean audible_bell)
@@ -1129,6 +1196,7 @@ update_visual_bell (gboolean visual_bell, gboolean audible_bell)
 
   return has_changed;
 }
+#endif /* HAVE_GCONF */
 
 #ifdef HAVE_GCONF
 static gboolean
@@ -1496,6 +1564,9 @@ meta_preference_to_string (MetaPreference pref)
     case META_PREF_FOCUS_MODE:
       return "FOCUS_MODE";
 
+    case META_PREF_RAISE_ON_CLICK:
+      return "RAISE_ON_CLICK";
+      
     case META_PREF_THEME:
       return "THEME";
 
@@ -1534,36 +1605,30 @@ meta_preference_to_string (MetaPreference pref)
 
     case META_PREF_BUTTON_LAYOUT:
       return "BUTTON_LAYOUT";
-      break;
 
     case META_PREF_WORKSPACE_NAMES:
       return "WORKSPACE_NAMES";
-      break;
 
     case META_PREF_VISUAL_BELL:
       return "VISUAL_BELL";
-      break;
 
     case META_PREF_AUDIBLE_BELL:
       return "AUDIBLE_BELL";
-      break;
 
     case META_PREF_VISUAL_BELL_TYPE:
       return "VISUAL_BELL_TYPE";
-      break;
 
     case META_PREF_REDUCED_RESOURCES:
       return "REDUCED_RESOURCES";
-      break;
+
     case META_PREF_GNOME_ACCESSIBILITY:
       return "GNOME_ACCESSIBILTY";
-      break;
+
     case META_PREF_CURSOR_THEME:
       return "CURSOR_THEME";
-      break;
+
     case META_PREF_CURSOR_SIZE:
       return "CURSOR_SIZE";
-      break;
     }
 
   return "(unknown)";
@@ -1704,6 +1769,52 @@ static MetaKeyPref window_bindings[] = {
   { NULL, 0, 0, FALSE }
 };
 
+#ifndef HAVE_GCONF
+typedef struct
+{
+  const char *name;
+  const char *keybinding;
+} MetaSimpleKeyMapping;
+
+/* Name field must occur in the same order as screen_bindings, though entries
+ * can be skipped
+ */
+static MetaSimpleKeyMapping screen_string_bindings[] = {
+  { META_KEYBINDING_WORKSPACE_LEFT,         "<Control><Alt>Left"         },
+  { META_KEYBINDING_WORKSPACE_RIGHT,        "<Control><Alt>Right"        },
+  { META_KEYBINDING_WORKSPACE_UP,           "<Control><Alt>Up"           },
+  { META_KEYBINDING_WORKSPACE_DOWN,         "<Control><Alt>Down"         },
+  { META_KEYBINDING_SWITCH_WINDOWS,         "<Alt>Tab"                   },
+  { META_KEYBINDING_SWITCH_PANELS,          "<Control><Alt>Tab"          },
+  { META_KEYBINDING_CYCLE_WINDOWS,          "<Alt>Escape"                },
+  { META_KEYBINDING_CYCLE_PANELS,           "<Control><Alt>Escape"       },
+  { META_KEYBINDING_SHOW_DESKTOP,           "<Control><Alt>d"            },
+  { META_KEYBINDING_PANEL_MAIN_MENU,        "<Alt>F1"                    },
+  { META_KEYBINDING_PANEL_RUN_DIALOG,       "<Alt>F2"                    },
+  { META_KEYBINDING_COMMAND_SCREENSHOT,     "Print"                      },
+  { META_KEYBINDING_COMMAND_WIN_SCREENSHOT, "<Alt>Print"                 },
+  { NULL,                                   NULL                         }
+};
+
+/* Name field must occur in the same order as window_bindings, though entries
+ * can be skipped
+ */
+static MetaSimpleKeyMapping window_string_bindings[] = {
+  { META_KEYBINDING_WINDOW_MENU,            "<Alt>Print"                 },
+  { META_KEYBINDING_MAXIMIZE,               "<Alt>F10"                   },
+  { META_KEYBINDING_UNMAXIMIZE,             "<Alt>F5"                    },
+  { META_KEYBINDING_MINIMIZE,               "<Alt>F9"                    },
+  { META_KEYBINDING_CLOSE,                  "<Alt>F4"                    },
+  { META_KEYBINDING_BEGIN_MOVE,             "<Alt>F7"                    },
+  { META_KEYBINDING_BEGIN_RESIZE,           "<Alt>F8"                    },
+  { META_KEYBINDING_MOVE_WORKSPACE_LEFT,    "<Control><Shift><Alt>Left"  },
+  { META_KEYBINDING_MOVE_WORKSPACE_RIGHT,   "<Control><Shift><Alt>Right" },
+  { META_KEYBINDING_MOVE_WORKSPACE_UP,      "<Control><Shift><Alt>Up"    },
+  { META_KEYBINDING_MOVE_WORKSPACE_DOWN,    "<Control><Shift><Alt>Down"  },
+  { NULL,                                   NULL                         }
+};
+#endif /* NOT HAVE_GCONF */
+
 static void
 init_bindings (void)
 {
@@ -1750,6 +1861,42 @@ init_bindings (void)
       g_free (str_val);      
       g_free (key);
 
+      ++i;
+    }
+#else /* HAVE_GCONF */
+  int i = 0;
+  int which = 0;
+  while (window_string_bindings[i].name)
+    {
+      /* Find which window_bindings entry this window_string_bindings entry
+       * corresponds to.
+       */
+      while (strcmp(window_bindings[which].name, 
+                    window_string_bindings[i].name) != 0)
+        which++;
+
+      /* Set the binding */
+      update_binding (&window_bindings[which],
+                      window_string_bindings[i].keybinding);
+
+      ++i;
+    }
+
+  i = 0;
+  which = 0;
+  while (screen_string_bindings[i].name)
+    {
+      /* Find which window_bindings entry this window_string_bindings entry
+       * corresponds to.
+       */
+      while (strcmp(screen_bindings[which].name, 
+                    screen_string_bindings[i].name) != 0)
+        which++;
+
+      /* Set the binding */
+      update_binding (&screen_bindings[which], 
+                      screen_string_bindings[i].keybinding);
+
       ++i;
     }
 #endif /* HAVE_GCONF */
@@ -1781,6 +1928,10 @@ init_commands (void)
 
       ++i;
     }
+#else
+  int i;
+  for (i = 0; i < MAX_COMMANDS; i++)
+    commands[i] = NULL;
 #endif /* HAVE_GCONF */
 }
 
@@ -1812,10 +1963,16 @@ init_workspace_names (void)
 
       ++i;
     }
+#else
+  int i;
+  for (i = 0; i < MAX_REASONABLE_WORKSPACES; i++)
+    workspace_names[i] = g_strdup_printf (_("Workspace %d"), i + 1);
+
+  meta_topic (META_DEBUG_PREFS,
+              "Initialized workspace names\n");
 #endif /* HAVE_GCONF */
 }
 
-#ifdef HAVE_GCONF
 static gboolean
 update_binding (MetaKeyPref *binding,
                 const char  *value)
@@ -1863,6 +2020,7 @@ update_binding (MetaKeyPref *binding,
   return changed;
 }
 
+#ifdef HAVE_GCONF
 static const gchar*
 relative_key (const gchar* key)
 {
@@ -2197,6 +2355,9 @@ meta_prefs_change_workspace_name (int         i,
     }
   
   g_free (key);
+#else
+  g_free (workspace_names[i]);
+  workspace_names[i] = g_strdup (name);
 #endif /* HAVE_GCONF */
 }
 
